<?php

namespace Teknisa\FIN\Modules\ContaPagar\Services;

use Doctrine\ORM\EntityManager;
use Teknisa\Libs\Util\NewCode;
use Teknisa\Libs\Util\DateUtil;
use Teknisa\Model\Geral\Entities\Paramgeral;
use Teknisa\Model\Financeiro\Entities\Boletpagar;
use Teknisa\Model\Financeiro\Entities\ContaPagar;
use Teknisa\Model\Financeiro\Entities\Tipoctpagar;
use Teknisa\Model\Financeiro\Entities\Ajctpagar;
use Teknisa\Model\Financeiro\Entities\Ajbolpag;
use Teknisa\Model\Financeiro\Entities\Tipoajpagar;
use Teknisa\Model\Financeiro\Entities\Movfluxo;
use Teknisa\FIN\Modules\FluxoCaixa\Services\FluxoDeCaixaService;
use Teknisa\FIN\Modules\ContaPagar\Factory\AssociacaoTituloFactory;

/**
 * Class AssociacaoTituloService
 * @package Teknisa\FIN\Modules\ContaPagar\Services
 */
class AssociacaoTituloService {

    const TAMANHO_NRLANBOLPAG       = 10;
    const STATUS_BOLETO_PENDENTE    = 'A';
    const ID_CNAB_N_ENVIADO         = '00';
    const NUMERO_PARCELAMENTO_BOL   = '00';
    const PREFIXO_NOVOCODIGO_BOLPAG = 'BOLETPAGAR';

    protected $entityManager;
    protected $associacaoTituloFactory;
    protected $fluxoDeCaixaService;
    protected $newCode;

    /**
     * @param EntityManager $entityManager
     */
    public function __construct(
        EntityManager $entityManager, AssociacaoTituloFactory $associacaoTituloFactory, FluxoDeCaixaService $fluxoDeCaixaService,
        NewCode $newCode
    ){
        $this->entityManager = $entityManager;
        $this->associacaoTituloFactory = $associacaoTituloFactory;
        $this->fluxoDeCaixaService = $fluxoDeCaixaService;
        $this->newCode = $newCode;
    }

    /**
     * Valida Associação Título
     * @param $boletos
     * @throws \Exception
     */
    public function validaAssociacaoTitulo($boletos)
    {
        try{
            // Variáveis de Inicialização
            $cdforneced  = null;
            $idtpctpagar = null;

            // Percorre Boletos a Pagar
            foreach ($boletos as $linha) {

                // Busca Boleto a Pagar
                /** @var $boletpagar Boletpagar */
                $boletpagar = $this->associacaoTituloFactory->buscaBoletPagar(
                    $linha['NRORG'], $linha['CDFILIAL'], $linha['NRLANBOLPAG']
                );

                // Verifica Existência Fornecedor
                if (!$cdforneced)
                    $cdforneced = $boletpagar->getCdforneced();

                // Verifica Diferença Fornecedor
                if ($boletpagar->getCdforneced() != $cdforneced)
                    throw new \Exception("Operação bloqueada. Apenas títulos de mesmo fornecedor podem ser associados.");

                // Busca Conta(s) a Pagar
                /** @var Contapagar $contaspagar */
                $contaspagar = $this->associacaoTituloFactory->buscaContaspagar($linha['NRORG'], $linha['CDFILIAL'], $linha['NRLANBOLPAG']);

                /** @var Contapagar $contapagar */
                foreach ($contaspagar as $contapagar) {
                    // Busca Tipo de Conta
                    /** @var $tipodeconta Tipoctpagar */
                    $tipodeconta = $this->associacaoTituloFactory->buscaTipoctpagar(
                        $contapagar->getNrorg(), $contapagar->getCdtpctpagar()
                    );

                    // Verifica Existência Tipo de Conta
                    if (!$idtpctpagar)
                        $idtpctpagar = $tipodeconta->getIdtpctpagar();

                    // Verifica Diferença Tipo de Conta [Normal/Adiantamento]
                    if ($idtpctpagar != $tipodeconta->getIdtpctpagar())
                        throw new \Exception("Operação bloqueada. Não é possível associar títulos de tipo de conta 'Adiantamento' com títulos de tipo de conta 'Normal'.");
                }

                // Verifica Título Enviado CNAB
                if (!in_array($boletpagar->getIdsitucnabp(),array('00','01')))
                    throw new \Exception("Operação bloqueada. Título enviado via CNAB não pode ser associado.");

                // Verifica Título Parcelado
                if ((bool)$boletpagar->getCdmotiparc() === true)
                    throw new \Exception("Operação bloqueada. Título parcelado não pode ser associado.");
            }
        }
        catch(\Exception $e){
            throw $e;
        }
    }

    /**
     * Gera Associação Titulo
     * @param $nrorg
     * @param $cdfilial
     * @param $boletos
     * @param $dtvenbolpag
     * @param $cdreducont
     * @param $cdtpbaixapag
     * @throws \Exception
     */
    public function geraAssociacaoTitulo($nrorg, $cdfilial, $boletos, $dtvenbolpag, $cdreducont, $cdtpbaixapag)
    {
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try {

            //Variaveis de Inicialização
            $vrfinalbpag = 0;

            // Formata Data de Vencimento
            $dtvenbolpag = $this->formataData($dtvenbolpag);

            // Gera Novo Código Boleto
            $nrlanbolpag = $this->geraNovoCodigo($nrorg, self::PREFIXO_NOVOCODIGO_BOLPAG.$cdfilial);

            // Percorre Boletos
            foreach ($boletos as $boleto) {
                // Calcula Valor Total Boletos
                $vrfinalbpag += $boleto['VRFINALBPAG'];
            }

            // Busca Boleto Pagar
            /** @var $boleto Boletpagar */
            $boleto =  $this->associacaoTituloFactory->buscaBoletpagar(
                $nrorg, $boletos[0]['CDFILIAL'], $boletos[0]['NRLANBOLPAG']
            );
            // Insere Boleto Pagar Associado
            /** @var Boletpagar $boletonovo */
            $boletonovo = $this->associacaoTituloFactory->insereBoletpagar(
                $nrorg, $nrlanbolpag, $cdfilial, $vrfinalbpag, $vrfinalbpag, $dtvenbolpag, $cdreducont, $cdtpbaixapag, $boleto->getCdforneced(),
                $boleto->getCdempresa(), self::STATUS_BOLETO_PENDENTE, self::ID_CNAB_N_ENVIADO, self::NUMERO_PARCELAMENTO_BOL, null, null,
                null, null, null, null, null, $boleto->getCdoperinclusao()
            );
            
            $this->entityManager->persist($boletonovo);
            $this->entityManager->flush();

            // Percorre Boletos a Pagar
            foreach ($boletos as $boleto){

                // Busca as Contas que compõe o boleto
                $contaspagar  = $this->associacaoTituloFactory->buscaContaspagar(
                    $nrorg, $boleto['CDFILIAL'], $boleto['NRLANBOLPAG']
                );

                // Percorre Contas a Pagar
                /** @var $conta Contapagar*/
                foreach ($contaspagar as $conta){
                    if((DateUtil::comparaDatas($this->formataData($conta->getDtorigvenpag()),$dtvenbolpag, 'IGUAL') == false)
                        or (DateUtil::comparaDatas($this->formataData($conta->getDtatuavenpag()),$dtvenbolpag, 'IGUAL') == false))
                    {
                        //Subtrai o valor do fluxo da conta na data anterior
                        /** @var Movfluxo $movfluxo */
                        $movfluxo = $this->fluxoDeCaixaService->atualizaFluxoCaixa(
                            $nrorg, $conta->getCdempresa(), $conta->getCdfimovflpag(), $conta->getCdclassfina(),
                            $conta->getDtmovfluxo(), $conta->getVrorigpag()*-1, $conta->getIdmovfluxo(), 'S'
                        );
                        $this->entityManager->persist($movfluxo);
                        $this->entityManager->flush();

                        //Acresce o valor do fluxo da conta na nova data de vencimento
                        /** @var Movfluxo $movfluxo */
                        $movfluxo = $this->fluxoDeCaixaService->atualizaFluxoCaixa(
                            $nrorg, $conta->getCdempresa(),$conta->getCdfimovflpag(),$conta->getCdclassfina(),
                            $dtvenbolpag,$conta->getVrorigpag(),$conta->getIdmovfluxo(), 'S'
                        );
                        $this->entityManager->persist($movfluxo);
                        $this->entityManager->flush();

                        // Atualiza Conta a Pagar
                        /** @var Contapagar $conta */
                        $conta = $this->associacaoTituloFactory->atualizaContapagar(
                            $conta, $nrlanbolpag, $cdfilial, $dtvenbolpag, $dtvenbolpag, null
                        );
                        $this->entityManager->persist($conta);

                        // Busca Ajuste(s) Conta a Pagar
                        /** @var Ajctpagar $ajustes */
                        $ajustes = $this->associacaoTituloFactory->buscaAjsctpagar(
                            $conta->getNrorg(), $conta->getCdfilial(), $conta->getNrlanctonf(),
                            $conta->getNrseqvenpag(), $conta->getNrparcelapag()
                        );
                        
                        // Percorre Ajuste(s) Conta Pagar
                        /** @var  $ajuste Ajctpagar*/
                        foreach ($ajustes as $ajuste)
                        {
                            // Busca Tipo de Ajuste a Pagar
                            /** @var  $tipoajuste Tipoajpagar */
                            $tipoajuste = $this->associacaoTituloFactory->buscaTipoajpagar($ajuste->getNrorg(), $ajuste->getCdajuctpag());

                            if($tipoajuste->getIdajuctpag() == 'D'){
                                /** @var Movfluxo $movfluxo */
                                $movfluxo = $this->fluxoDeCaixaService->atualizaFluxoCaixa(
                                    $nrorg, $ajuste->getCdemmfajcntp(), $ajuste->getCdfimfajcntp(), $ajuste->getCdclassfina(),
                                    $ajuste->getDtmovfluxo(), $ajuste->getVrajustepag(), 'E', 'S'
                                );
                                $this->entityManager->persist($movfluxo);
                                $this->entityManager->flush();

                                /** @var Movfluxo $movfluxo */
                                $movfluxo = $this->fluxoDeCaixaService->atualizaFluxoCaixa(
                                    $nrorg, $ajuste->getCdemmfajcntp(), $ajuste->getCdfimfajcntp(), $ajuste->getCdclassfina(),
                                    $dtvenbolpag, $ajuste->getVrajustepag()*-1, 'E', 'S'
                                );
                                $this->entityManager->persist($movfluxo);
                                $this->entityManager->flush();
                            }else{
                                /** @var Movfluxo $movfluxo */
                                $movfluxo = $this->fluxoDeCaixaService->atualizaFluxoCaixa(
                                    $nrorg, $ajuste->getCdemmfajcntp(), $ajuste->getCdfimfajcntp(), $ajuste->getCdclassfina(),
                                    $ajuste->getDtmovfluxo(), $ajuste->getVrajustepag()*-1, 'E', 'S'
                                );
                                $this->entityManager->persist($movfluxo);
                                $this->entityManager->flush();

                                /** @var Movfluxo $movfluxo */
                                $movfluxo = $this->fluxoDeCaixaService->atualizaFluxoCaixa(
                                    $nrorg, $ajuste->getCdemmfajcntp(), $ajuste->getCdfimfajcntp(), $ajuste->getCdclassfina(),
                                    $dtvenbolpag, $ajuste->getVrajustepag(), 'E', 'S'
                                );
                                $this->entityManager->persist($movfluxo);
                                $this->entityManager->flush();

                            }
                            // Atualiza Ajuste Conta a Pagar
                            $ajuste->setDtmovfluxo($dtvenbolpag);

                            $this->entityManager->persist($ajuste);
                        }
                    }
                    else{
                        // Atualiza Conta a Pagar
                        $conta->setCdfilibpag($cdfilial);
                        $conta->setNrlanbolpag($nrlanbolpag);

                        $this->entityManager->persist($conta);
                    }
                }

                // Busca Ajustes Boleto a Pagar
                /** @var Ajbolpag $ajustes */
                $ajustes = $this->associacaoTituloFactory->buscaAjsbolpag(
                    $boleto['NRORG'], $boleto['CDFILIAL'], $boleto['NRLANBOLPAG']
                );
                
                if($ajustes) {
                    /** @var Ajbolpag $ajuste */
                    foreach ($ajustes as $ajuste){
                        
                        // Verifica se o ajuste já foi gravado no novo boleto
                        /** @var $ajusteaux Ajbolpag */
                        $ajusteaux = $this->associacaoTituloFactory->buscaAjbolpag(
                            $boleto['NRORG'], $cdfilial, $nrlanbolpag, $ajuste->getCdajuctpag()
                        );
                        
                        if (!$ajusteaux){ 
                            // Insere Ajuste Boleto a Pagar
                            /** @var Ajbolpag $ajusteaux */
                            $ajusteaux = $this->associacaoTituloFactory->insereAjbolpag(
                                $ajuste->getNrorg(), $cdfilial, $nrlanbolpag, $ajuste->getCdajuctpag(), $ajuste->getVrajbolpag()
                            );
                        }else{
                            // Atualiza Valor Ajuste Boleto a Pagar
                            $ajusteaux->setVrajbolpag($ajusteaux->getVrajbolpag()+$ajuste->getVrajbolpag());
                        }
                        
                        $this->entityManager->persist($ajusteaux);
                        $this->entityManager->remove($ajuste);
                        $this->entityManager->flush();
                    }
                }

                // Busca Boleto a Pagar
                /** @var Boletpagar $boletpagar */
                $boletpagar = $this->associacaoTituloFactory->buscaBoletpagar(
                    $nrorg,$boleto['CDFILIAL'],$boleto['NRLANBOLPAG']
                );

                // Deleta Boleto a Pagar
                $this->entityManager->remove($boletpagar);
            }
            $this->entityManager->flush();
            $connection->commit();
        }catch (\Exception $e) {
            $connection->rollback();
            throw $e;
        }
    }

    /**
     * @param $boletos
     * @throws \Exception
     */
    public function cancelaAssociacaoTitulo($boletos){
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try {
            // Percorre Boletos a Pagar
            foreach ($boletos as $boleto) {

                // Busca Boleto a Pagar(Associado)
                /** @var Boletpagar $boletpagar */
                $boletpagar_assoc = $this->associacaoTituloFactory->buscaBoletpagar(
                    $boleto['NRORG'], $boleto['CDFILIAL'], $boleto['NRLANBOLPAG']
                );

                //Busca Conta(s) Boleto a Pagar(Associado)
                $contas = $this->associacaoTituloFactory->buscaContaspagar(
                    $boletpagar_assoc->getNrorg(), $boletpagar_assoc->getCdfilial(), $boletpagar_assoc->getNrlanbolpag()
                );

                // Percorre Conta(s) Boleto a Pagar(Associado)
                /** @var $conta Contapagar */
                foreach ($contas as $conta) {

                    // Insere Boleto a Pagar
                    $boletpagar = $this->geraBoletoAPagar(
                        $conta->getCdfilibpag(), $boletpagar_assoc->getNrorg(), $boletpagar_assoc->getCdforneced(),
                        $boletpagar_assoc->getCdempresa(), $conta->getVratuapag(), $boletpagar_assoc->getVrpemulbpag(),
                        $boletpagar_assoc->getVrpemorbpag(), $boletpagar_assoc->getDtvenbolpag(), $boletpagar_assoc->getCdreducont(),
                        $boletpagar_assoc->getCdbancbolp(), $boletpagar_assoc->getCdtpbaixapag(), $boletpagar_assoc->getCdagenbolp(),
                        $boletpagar_assoc->getCdccorbolp(), $boletpagar_assoc->getCdoperinclusao()
                    );
                    $this->entityManager->persist($boletpagar);
                    $this->entityManager->flush();

                    // Atualiza Conta a Pagar
                    $conta->setCdfilibpag($boletpagar->getCdfilial());
                    $conta->setNrlanbolpag($boletpagar->getNrlanbolpag());
                    $this->entityManager->persist($conta);
                }

                // Busca Ajuste Boleto a Pagar
                /** @var Ajbolpag $ajustes */
                $ajsbolpag = $this->associacaoTituloFactory->buscaAjsbolpag(
                    $boletpagar_assoc->getNrorg(), $boletpagar_assoc->getCdfilial(), $boletpagar_assoc->getNrlanbolpag()
                );

                // Verifica Existência Ajuste(s) Boleto a Pagar(Associado)
                if($ajsbolpag)
                    // Percorrer Ajuste(s) Boleto a Pagar(Associado)
                    /** @var $ajuste Ajbolpag */
                    foreach ($ajsbolpag as $ajbolpag)
                        // Deleta Ajuste Boleto a Pagar(Associado)
                        $this->entityManager->remove($ajbolpag);

                $this->entityManager->flush();

                // Deleta Boleto a Pagar(Associado)
                $this->entityManager->remove($boletpagar_assoc);
            }
            $this->entityManager->flush();
            $connection->commit();
        }catch (\Exception $e) {
            $connection->rollback();
            throw $e;
        }
    }

    /**
     * @param $cdfilial
     * @param $nrorg
     * @param $cdforneced
     * @param $cdempresa
     * @param $vrbolpag
     * @param $vrpemulbpag
     * @param $vrpemorbpag
     * @param $dtvenbolpag
     * @param $cdreducont
     * @param $cdbancbolp
     * @param $cdtpbaixapag
     * @param $cdagenbolp
     * @param $cdccorbolp
     * @return null|Boletpagar
     * @throws \Exception
     */
    public function geraBoletoAPagar(
        $cdfilial,$nrorg, $cdforneced, $cdempresa, $vrbolpag, $vrpemulbpag, $vrpemorbpag,
        $dtvenbolpag, $cdreducont, $cdbancbolp, $cdtpbaixapag, $cdagenbolp, $cdccorbolp, $cdoperinclusao
    ){
        try {
            // Gera Novo Código Boleto a Pagar
            $nrlanbolpag = $this->geraNovoCodigo($nrorg,'BOLETPAGAR'.$cdfilial);

            // Formata Data de Vencimento
            $dtvenbolpag = $this->formataData($dtvenbolpag);

            // Calcula Valores Boleto a Pagar
            $vrmultabpag  = $vrbolpag*($vrpemulbpag/100);
            $vrmordiabpag = $vrbolpag*($vrpemorbpag/100);

            // Insere Boleto a Pagar
            $boletpagar = $this->associacaoTituloFactory->insereBoletpagar(
                $nrorg, $nrlanbolpag, $cdfilial, $vrbolpag, $vrbolpag, $dtvenbolpag, $cdreducont, $cdtpbaixapag,
                $cdforneced, $cdempresa, self::STATUS_BOLETO_PENDENTE, self::ID_CNAB_N_ENVIADO, self::NUMERO_PARCELAMENTO_BOL,
                $vrpemulbpag, $vrmultabpag, $vrpemorbpag, $vrmordiabpag, $cdbancbolp, $cdagenbolp, $cdccorbolp, $cdoperinclusao
            );

            return $boletpagar;
        } catch (\Exception $e) {
            throw $e;
            return null;
        }
    }

    /**
     * Gera Novo Código
     * @param $cdcontador
     * @return bool|null
     * @throws \Exception
     */
    public function geraNovoCodigo(
        $nrorg, $cdcontador
    ){
        try {

            // Gera Novo Código Boleto
            $nrlanbolpag = $this->newCode->getCodeByTableName($cdcontador, $nrorg);

            return $nrlanbolpag;
        } catch (\Exception $e) {
            throw $e;
        }
    }

    /**
     * Formata Data
     * @param $data
     * @return \DateTime|null|string|\Zeedhi\Framework\ORM\DateTime
     * @throws \Exception
     */
    public function formataData($data){
        if (!$data)
            return null;

        if (strlen($data) == 19)
            $data = DateUtil::truncateString($data);
        $data = DateUtil::getDataDeString($data);
        $data = DateUtil::truncateData($data);

        return $data;
    }

}